//JA50_IRsense.ino

char thisrobot[] ="Jet";
char thisprog[]  ="JA50_IRsense";
/*
Sketch uses 28062 bytes (86%) of program storage space. Maximum is 32256 bytes.
Global variables use 1293 bytes (63%) of dynamic memory, leaving 755 bytes for local variables. Maximum is 2048 bytes.
*/
//set Serial terminal to 9600 baud
//== #define ================================================
//#define BambinoPM         //ProMini 328,5v
//#define BambinoMM         //Uno with no Whisker sensors
//#define BambinoM2         //Uno 
#define Jet               //Uno    affects JApins.h -  Edit manually JAmovement.ino, JAvarsCons.h
//#define servolib          //       affects JAbehaviour48, JAmovement48.ino
//#define baud 19200        //ZumCore
#define baud 9600               
/*
== JA changes from Bambino ==================================
JApins48.h   Jet pin allocations
commands bp,bt,br, scale factor changed from 2 for Bambino to 4 
JAmovement48.ino same as movement34.ino with added routine for servolib
== Notes ====================================================
  
Introducing

Jet now fitted with 4way IR obstacle sensors, on each foot and two on the body
plus an IR line/drop-off sensor on each foot.
    'bwf' - set stepping direction forwards  in L and R, used in Reactions 
    'bwb' - set stepping direction backwards in L and R, used in Reactions
============================================================
Note - The Arduino Serial Monitor can't send [Esc]. 
       If another Terminal is used which can send [Esc] then at GetCcmnd() this will be printed,
	     which may print as a symbol or the Terminal may change it to "."
     
Bugs -
============================================================
/*History 
 * 2021-04-07  added && Ccmnd!='0'   //if just got '0' for Acts0 then ...... 2021-04-07
 *             { if (f_bAutoSense && !f_reacting  && Ccmnd!='0')   //don't test if in a React - no return stack 
 *                   //if just got '0' for Acts0 then Reacting looses epage before EEPROMindex is set - 2021-04-07
 * 2021-04-06 queryEdgeLR() ORs holes and toes
 * 2021-03-30 changed round rgroup and rname so rname is 0-9 and rgroup is A-Z
 *            made rname sub to rgroup, brought JA49 and JA48 into line and Bambino from Name onwards.
 * 2021-03-29 '9' - do default act
 * 2021-03-22 '?$'
 * JA50
*/
/*== NEW ====================================================
Doc_JA59.h           text removed from here
Doc_JASections.h     text removed from here
Doc_JAcmnds50.h      text removed from here
/*== CHANGED ================================================
system ? case '$':  readUSdistance(); break;
        -> case '$':  readUSdistance(); printIRsensors();  break;
void printBehaviour() - added command codes for each entry
*/
//== INCLUDES ===============================================
#include "JApins48.h"
#include "JAvarsCons49.h"
#include <EEPROM.h>           // to read and write internal EEPROM
  // EEPROM.write(address,bytevalue)
  // bytevalue =EEPROM.read(address)
#include <SoftwareSerial.h>   // for IR commands from PicAxe
  SoftwareSerial IRserial(2, 4); // RX, TX, TX not used
#ifdef servolib
  #include <Servo.h>  
    Servo servoRoll;
    Servo servoPace;
 #endif    

//== VARIABLES and CONSTANTS ================================
//see also TAB FILE > varsCons49.h
boolean f_recording =false;   //recording an Act
boolean f_recordC =false;     //record current Command
boolean f_play =false;        //Play current Act
boolean f_reacting =false;    //in a ReAct - set in behaviour44.ino if a React is initiated.
boolean f_inDo =false;
boolean f_runDefault =false;  //run Default when nothing else to do
boolean f_skipmind =false;    //so commands in a mind which are skipped do not print in GetCmnd
const byte NL =10;            //NewLine value
const byte LF =10;            //LineFeed value                                  
const byte CR =13;            //Carriage Return value
const byte Esc =27;           //Esc key value
char Act;                     //current Act as '1' - '8'
byte actIndex =0;             //current Act number, set to default           
int EEPROMindex =-1;          //index to moves in an Act in EEPROM, -1 => nothing to continue with #>
int hashhashAt;               //copy of EEPROMindex at hashhashSTOP 
int ReactAt;                  //copy of EEPROMindex at React
int ActEnd;                   //end address of current recording Act in EEPROM
int DoAt;                     //return address in a Do
int moves =0;                 //current move
char Ccmnd =0;                //current Character command, default null command
byte epage =1;                //the Act page for the current Act, default page =1 for  Acts 1-8 and page =2 for 01-08
char Behaviour;               //read from EEPROM in init, current behaviour 1-8 in epage 0
byte BehaviourEbase =224;     //base address of current behaviour 0,32,64,96,128,160,192,(224)
byte slotbase;                //working slotbase
byte behaveNum;               //BeHave address in Behaviour slot 
byte behaveValue;
String numString ="";         //temporary string in loadBehaviour(), getnum()
byte mchance;                 //chance of executing Mind cmnds   
//== SETUP ==================================================
void setup() 
{ 
  f_BeQuiet =0;                    // send status messages
  f_MsgForMe =1;                   // listen to messages   
  f_hashhashSTOP =0;               
  
  Serial.begin(baud);      // to Monitor, baud for ZumCore 
  Serial.println();
  Serial.print(thisprog);  // so we know which program is running
  Serial.print(" ");
#ifdef servolib
  Serial.println(" servolib");
#endif  

  IRserial.begin(9600);     // set the data rate for the SoftwareSerial port
     
  pinMode(voicePin,OUTPUT);
  pinMode(LbeardLEDpin_, OUTPUT);  
  pinMode(RbeardLEDpin_, OUTPUT);  
  pinMode(LwhiskerLEDpin_, OUTPUT);
  pinMode(RwhiskerLEDpin_, OUTPUT);
  digitalWrite(LbeardLEDpin_, HIGH);  // beard LED OFF
  digitalWrite(RbeardLEDpin_, HIGH);  // beard LED OFF
  digitalWrite(LwhiskerLEDpin_, HIGH);// whisker LED OFF
  digitalWrite(RwhiskerLEDpin_, HIGH);// whisker LED OFF    
  
  if (EEPROM.read(E8f_walking)>1) {initialiseEEPROM();}   // in case it is not initialised
  
  //find current current behaviour in EEPROM and copy to RAM
  for (BehaviourEbase =0; BehaviourEbase < 224; BehaviourEbase +=32 )  //test first 7 slots 
    {if (EEPROM.read(BehaviourEbase) =='@')  break; }
  if (EEPROM.read(BehaviourEbase) !='@') { EEPROM.write(BehaviourEbase,'@'); }    //if '@' not found at 8 make 8 the default 
  Behaviour =BehaviourEbase /32 +'1';  //'1' - '8'
  getMind();
  Serial.print(thisrobot); // so we know which robot the program is for
  Serial.print(" ");
  Serial.print(rgroup);    // so we know which group it is in
  Serial.println(rname);   // so we know which robot this is
    
  //now configure Bambino
  f_forwards =1;    // set default

  pinMode(servoRollPin, OUTPUT);    // sets the digital pin as output
  pinMode(servoPacePin, OUTPUT);    // sets the digital pin as output
  digitalWrite(servoRollPin, LOW);   // initialise pin
  digitalWrite(servoPacePin, LOW);   // initialise pin

  if (f_walking) { Serial.println(" walking ON"); } else { Serial.println(" walking OFF"); }

#ifdef servolib
#else
  ServosInitC();  //initialises rollat to rollC and paceat to paceC
#endif

  rollat =rollC;
  paceat =paceC;   
  Srollat =rollat;
  Spaceat =paceat; 

#ifdef servolib
  servoRoll.write(rollat);
  servoPace.write(paceat);
  if (f_walking)  
  { servoRoll.attach(6);
    servoPace.attach(7);
  } 
 #endif 
}
//== LOOP ===================================================
void loop()
{ 
  f_recordC =true;     // record Command, default =true
  
#ifdef BambinoMM 
  redLEDdim();         //indicate ready for command except    
  if (f_recording)  { redLEDon(); }
  if (f_play)       { redLEDoff(); }
#else
  if (f_recording)  { redLEDon(); } else { redLEDoff(); }
  if (!f_reacting)                    // don't change if in a ReAct
    { if (f_play)
        { whiskerLEDsOff(); }
        else
        { whiskerLEDsDim(); }      //indicate ready for command
    }
#endif

  if (f_doDefault)   // true on the first loop and #< only
    { f_doDefault =false;                // don't do it again
      Ccmnd =autoAct;
      if(Ccmnd=='0') { f_runDefault =false;  Ccmnd ='.'; }     // '0' -> no autorun, can't run it so do stop 
    }   
   else
    // if f_hashhashSTOP skip GetCcmnd() and process Ccmnd '.' from ##
    { if (!f_hashhashSTOP)
        { if (f_bAutoSense && !f_reacting  && Ccmnd!='0')   //don't test if in a React - no return stack 
              //if just got '0' for Acts0 then Reacting looses epage before EEPROMindex is set - 2021-04-07
            { if (f_bG) { BeardReact(); }
              if (f_bW) { WhiskerReact(); }
              if (f_bU) { USreact(); }
            }
          if (!f_play && !f_reacting && !f_BeQuiet)  { Serial.print(">"); }    //ready prompt
          GetCcmnd();    //reads Serial or playing Act until Ccmnd != 0, prints Ccmnd
        }  
    }

DoCommand:
  switch (Ccmnd) 
  { case LF:    f_recordC =false;  break;     //linefeed
    case CR:    f_recordC =false;  break;     //carriage return
    case ' ':   break;     //strip space
    case ',':   break;     //strip comma
    case 'Q':   // fall through
    case Esc:   f_reacting =false; f_inDo =false; DoAt =0;  //fall through to '.'
    case '.':   if (f_recording) {  EEPROM.write(EEPROMindex,'.'); }     //terminated string
                f_recording =false;
                if (f_reacting)    // end of ReAct
                  { f_reacting =false; 
                    EEPROMindex =ReactAt; beardLEDsOff(); 
                    break;
                  }
                if (DoAt !=0) 
                  { Serial.print("("); Serial.print(DoAt); Serial.print(")"); 
                    EEPROMindex =DoAt; DoAt =0; f_inDo =true;
                    break;
                  }                               
                f_play =false;        // end of Act  
                if (!f_hashhashSTOP) { EEPROMindex =-1; };   // so there is nothing to continue with #>
                f_hashhashSTOP =false;
                if(!f_BeQuiet){ Serial.println(); }
                if (f_runDefault) { f_doDefault =true; }
                break;
    case '(':   f_inDo =true;   Serial.print("="); Serial.println(EEPROMindex-1); break;
    case ')':   f_inDo =false;  Serial.print("="); Serial.println(EEPROMindex-1); break;

    case 'A':	  // if recording record 'A' otherwise do beAlive() until 'Q' or reset
                if (!f_recording) { beAlive(); }  break;
    case 'X':	  // if recording record 'X' otherwise do Xplore() until 'Q' or reset
                if (!f_recording) { Xplore(); }  break; 
    case 'E':	  // if recording record 'E' otherwise do Experiment() until 'Q' or reset
                if (!f_recording) { Experiment(); }  break; 
    case '!':   epage =3;  break;  //reset to 1 at the end of setEEPROMindexes(char Act)
    case '0':   epage =2;  break;  //reset to 1 at the end of setEEPROMindexes(char Act)
    case '1':   // fall through
    case '2':   // fall through
    case '3':   // fall through
    case '4':   // fall through
    case '5':   // fall through
    case '6':   // fall through
    case '7':   // fall through
    case '8':   if (!f_recording) 
                  { if (f_inDo && DoAt==0)
                    { DoAt =EEPROMindex; Serial.print("(@"); Serial.print(DoAt); Serial.print(")"); 
                      f_inDo =false;        //because we are jumping to another Act
                    }  
                    f_play =true; Act =Ccmnd;  setEEPROMindexes(Act); 
                  } break;
    case '9':   f_hashhashSTOP =false;  f_doDefault =true;  break;
    case 'F':   forward();      break;        
    case 'B':   backward();     break;        
    case 'R':   turnRight();    break;        
    case 'L':   turnLeft();     break;        
    case 'H':   halt();         break; 
    case 'V':   whee();         break;   //Spkr and LED on different pins           
    case 'S':   f_recordC =false;
                toggle_f_walking(); 
                if(!f_BeQuiet)
                  {if (f_walking) {Serial.println(" Walking ON");} else {Serial.println(" Walking OFF");}}
                break;    
    case 'U':   USreact();      break; 
    case 'W':   WhiskerReact(); break; 
    case 'G':   BeardReact();   break;
voice:    
    case 'v':   if (f_recording) { RecordCcmnd(); }    //so 'v' is recorded
                GetCcmnd();                            //get parameter
                switch (Ccmnd) 
                { case 'b':  bip();   break;
                  case 'e':  whee();  break;
                  case 'o':  whoa();  break;
                  default:   bip();   Ccmnd ='b';  break;
                }    // end switch
                break;      
wait:
    case 'w':   if (f_recording) { RecordCcmnd(); }    //so 'w' is recorded
                GetCcmnd();                            //get parameter
                makeCcmndDigit();                      //ensure '0'-'9'
                waitCcmndSecs();
                break;                                 // if recording Ccmnd will be recorded
speed:                
    case 's':   if (f_recording) { RecordCcmnd(); }    //so 's' is recorded
                GetCcmnd();                            //get parameter
                if (Ccmnd=='u') { speedup();  break; }     // and if recording record 'u' 
                if (Ccmnd=='d') { speeddown();  break; }   // and if recording record 'd' 
                if (Ccmnd=='m') { rollspeedmax =rollspeed; pacespeedmax =pacespeed; break; }
                if (Ccmnd=='r') 
                { if (f_recording) { RecordCcmnd(); }  //so 'r' is recorded
                  GetCcmnd();                          //get parameter
                  if (Ccmnd=='m') { rollspeedmax =rollspeed; break; }
                  //else                 
                  makeCcmndDigit();                    //ensure '1'-'9'    
                  rollspeed =Ccmnd -'0';
                  break;                               // and if recording record digit
                }
                if (Ccmnd=='p') 
                { if (f_recording) { RecordCcmnd(); }  //so 'p' is recorded
                  GetCcmnd();                          //get parameter
                  if (Ccmnd=='m') { pacespeedmax =pacespeed; break; }
                  //else                 
                  makeCcmndDigit();                    //ensure '1'-'9'    
                  pacespeed =Ccmnd -'0';
                  break;                               // and if recording record digit 
                }
                //else                 
                makeCcmndDigit();                      //ensure '0'-'9'    
                rollspeed =max(1,Ccmnd -'0');          // hangs if 0 
                pacespeed =max(1,Ccmnd -'0');          // hangs if 0       	
                break;                                 // and if recording record digit 
turn:                
    case 't':   if (f_recording) { RecordCcmnd(); }    //so 't' is recorded
                GetCcmnd();                            //get parameter
                switch (Ccmnd)
                { case 'a':  turnAway();    break;
                  case 'm':  turnMemory();  break;
                  case 'o':  turnOther();   break;
                  case 'l': turnLight();   break;
                  case 'd': turnDark();    break;
                  case '@': if (f_recording) { RecordCcmnd(); }    //so '@' is recorded
                            GetCcmnd();                            //get parameter
                            switch (Ccmnd)
                            { case 'l':  turnAtLight();  break;
                              case 'd':  turnAtDark();   break;
                              default:   badkey();  if(!f_BeQuiet){Serial.println("bad key");}
                                         Ccmnd ='?';     //so if recording then '?' is recorded
                                         break;          //don't do anything 
                            }    // end switch
                  default:  turnMemory();   Ccmnd ='m';  break;
                }     // end switch case t
                break;
    case 'm':   if (f_recording) { RecordCcmnd(); }    //so 'm' is recorded
                GetCcmnd();                            //get parameter     
                mchance =random(0,10);
                //Serial.println(mchance);
                if (!f_recording && Ccmnd -'0' < mchance)        // don't skip if recording
                { f_skipmind =true; 
                  while(Ccmnd!='}') { GetCcmnd(); } 
                  f_skipmind =false; printCcmnd();     //so '}' is printed after skipping
                  }
                break;
    case '}':   break;  //end of a mind, ignore it

//---behaves----------------------------------------                
    case 'b':   if (f_recording) { RecordCcmnd(); }    //so 'b' is recorded
                GetCcmnd();                            //get parameter
                switch (Ccmnd)
                {
  b_behaviour:    case '0': if (f_recording) { RecordCcmnd(); }      //so Ccmnd is recorded
                            getMind();  break;
                  case '1': // fall through
                  case '2': // fall through
                  case '3': // fall through
                  case '4': // fall through
                  case '5': // fall through
                  case '6': // fall through
                  case '7': // fall through
                  case '8': if (f_recording) { RecordCcmnd(); }      //so Ccmnd is recorded
                            Behaviour =Ccmnd;
                            EEPROM.write(0 +BehaviourEbase,'.');     //deselect old behaviour 
                            BehaviourEbase =(Behaviour -'1') *32;    //new behaviour 
                            EEPROM.write(0 +BehaviourEbase,'@');     //new behaviour is current
                            getMind();
                            break;
  b_save:         case 's': if (f_recording) { RecordCcmnd(); }    //so 's' is recorded
                            GetCcmnd();                            //get parameter, if cmnd='0' then save to current behaviour  
                            if(Ccmnd<'0' || Ccmnd>'8') { Ccmnd ='?'; break; }
							if (f_recording) { RecordCcmnd(); }          //so Ccmnd is recorded
                            if(Ccmnd=='0') {slotbase =BehaviourEbase;} else {slotbase =(Ccmnd -'1') *32;}
                            saveMind();                 // save current behaviour to a behaviour slot
                            break;
  b_load:         case 'l': if (f_recording) 
                              { Ccmnd ='.'; RecordCcmnd();  //so '.' is recorded
                                f_recording =false;
                                break;
                               }
                            if (!f_recording) {loadBehaviour();}
                            break;	
  b_hash:         case '#': if (f_recording) { RecordCcmnd(); }    //so '#' is recorded
                            GetCcmnd();                     //get parameter
                            if (Ccmnd>='0' && Ccmnd<='9') 
                              { rgroup =Ccmnd;  break; }
                            if (Ccmnd>='A' && Ccmnd<='z') 
                              { rname =Ccmnd;  break; }
                            badkey();
                            break;
  b_autoRun:      case 'a': if (f_recording) { RecordCcmnd(); }    //so 'a' is recorded
                            GetCcmnd();                            //get parameter
                            if(Ccmnd<'0' || Ccmnd>'9') { Ccmnd ='0'; break; }
                            if (Ccmnd=='9') {f_runDefault =true; break;}          // ## sets f_runDefault false
                            autoAct =Ccmnd;
                            break;
  b_quiet:        case 'q': f_BeQuiet =1;  break; 
  b_walking:      case 'w': if (f_recording) { RecordCcmnd(); }    //so 'w' is recorded 
                            GetCcmnd();                            //get parameter
                            if (Ccmnd=='0') { disable_walking();  break; }   // and if recording record '0' 
                            if (Ccmnd=='1') { enable_walking();   break; }   // and if recording record '1' 
                            if (Ccmnd=='f') { f_forwards =true;   break; }   // and if recording record 'f' 
                            if (Ccmnd=='b') { f_forwards =false;  break; }   // and if recording record 'b' 
                            Ccmnd ='?';   // bad char, no action and if recording record '?' 
                            break;                                 // if recording Ccmnd will be recorded                                                       
  b_pace:         case 'p': if (f_recording) { RecordCcmnd(); }    //so 'p' is recorded
                            GetCcmnd();                            //get parameter
                            switch (Ccmnd)
    b_paceCplus:            { case '+': if (f_recording) { RecordCcmnd(); }    //so '+' is recorded
                                        GetCcmnd();                            //get parameter
                                        makeCcmndDigit();                      //ensure '0'-'9'
                                        paceC =min(EEPROM.read(BehaviourEbase +Es29paceC) +maxpace,paceC +(Ccmnd -'0'));  //restrict to maxpace  
                                        break;
    b_paceCminus:             case '-': if (f_recording) { RecordCcmnd(); }    //so '-' is recorded
                                        GetCcmnd();                            //get parameter
                                        makeCcmndDigit();                      //ensure '0'-'9'
                                        paceC =max(EEPROM.read(BehaviourEbase +Es29paceC) -maxpace,paceC -(Ccmnd -'0'));  //restrict to maxpace  
                                        break;
    b_paceby:                 default:  if (Ccmnd<'1' || Ccmnd>'9') 
                                        { badkey();  if(!f_BeQuiet){Serial.println("bad key"); }
                                          Ccmnd ='?';     //so if recording then '?' is recorded
                                          break; }
                                        paceby =(Ccmnd -'0') *pacebyScale;
                             }            
                            break;
  b_turnby:       case 't': if (f_recording) { RecordCcmnd(); }    //so 't' is recorded
                            GetCcmnd();                            //get parameter
                            makeCcmndDigit();                      //ensure '1'-'9'
                            turnby =(Ccmnd -'0') *turnbyScale;
                            break;
  b_roll:         case 'r':  if (f_recording) { RecordCcmnd(); }    //so 'r' is recorded
                            GetCcmnd();                            //get parameter
                            switch (Ccmnd)                            
    b_rollCplus:            { case '+': if (f_recording) { RecordCcmnd(); }    //so '+' is recorded
                                        GetCcmnd();                            //get parameter
                                        makeCcmndDigit();                      //ensure '0'-'9'
                                        rollC =min(EEPROM.read(BehaviourEbase +Es30rollC) +maxroll,rollC +(Ccmnd -'0'));  //restrict to maxroll
                                        break;
    b_rollCminus:             case '-': if (f_recording) { RecordCcmnd(); }    //so '-' is recorded
                                        GetCcmnd();                            //get parameter
                                        makeCcmndDigit();                      //ensure '0'-'9'
                                        rollC =max(EEPROM.read(BehaviourEbase +Es30rollC) -maxroll,rollC -(Ccmnd -'0'));  //restrict to maxroll
                                        break;
    b_rollby:                 default:  if (Ccmnd<'1' || Ccmnd>'9') 
                                        { badkey();  if(!f_BeQuiet){Serial.println("bad key"); }
                                          Ccmnd ='?';     //so if recording then '?' is recorded
                                          break; }
                                        rollby =(Ccmnd -'0') *rollbyScale;
                            }            
                            break;
  b_sonar:        case 'u': if (f_recording) { RecordCcmnd(); }    //so 'u' is recorded
                            GetCcmnd();                             //get parameter
                            switch (Ccmnd)
    b_sonarRange:           { case 'm': // don't record 'm'
                                        do
                                        { Serial.println("press [+] for AwareAt, [-] for TooNear, [.] or [Esc] to Quit");
                                          GetCcmnd();                            //get parameter
                                          objectRange =dist_HC_SR04(); if(f_cm) {objectRange /=5;} else {objectRange /=2;}
                                          if (Ccmnd=='+') 
                                            { SonarAwareAt =objectRange; 
                                              if (f_recording) { RecordCcmnd(); }    //so '+' is recorded
                                              Ccmnd ='0' +SonarAwareAt; if (f_recording) { RecordCcmnd(); }    //so 'SonarAwareAt' is recorded
                                              }
                                          if (Ccmnd=='-')
                                            { SonarTooNear =objectRange;
                                              if (f_recording) { RecordCcmnd(); }    //so '-' is recorded
                                              Ccmnd ='0' +SonarTooNear; if (f_recording) { RecordCcmnd(); }    //so 'SonarTooNear' is recorded
                                              }
                                        } while (Ccmnd!='.' &&  Ccmnd!='Q' && Ccmnd!=Esc);
                                        break;
    b_sonarAwareAt:           case '+': if (f_recording) { RecordCcmnd(); }    //so '+' is recorded
                                        GetCcmnd();                            //get parameter
                                        makeCcmndDigit();                      //ensure '1'-'9'
                                        if (f_recording) { RecordCcmnd(); }    //so Ccmnd is recorded
                                        SonarAwareAt =Ccmnd -'0';
                                        break;
    b_sonarTooNear:           case '-': if (f_recording) { RecordCcmnd(); }    //so '-' is recorded
                                        GetCcmnd();                            //get parameter
                                        makeCcmndDigit();                      //ensure '1'-'9'
                                        if (f_recording) { RecordCcmnd(); }    //so Ccmnd is recorded
                                        SonarTooNear =Ccmnd -'0';
                                        break;
    b_sonarUnits:             case 'c': if (f_recording) { RecordCcmnd(); }    //so 'c' is recorded
                                        GetCcmnd();                            //get parameter
                                        if((Ccmnd -'0') ==1) { f_cm =true; } else { f_cm =false; }
                                        break;
    b_sonarSave:              case 's': EEPROM.write(BehaviourEbase +Es19SonarAwareAt,SonarAwareAt);
                                        EEPROM.write(BehaviourEbase +Es18SonarTooNear,SonarTooNear);
                                        break;
                            }     // end switch
                            break;
  b_GroundReact:  case 'G': if (f_recording) { RecordCcmnd(); }    //so 'G' is recorded
                            GetCcmnd();                            //get parameter           
                            if (Ccmnd=='0') { f_bG =0; } 
                            else 
                            { Ccmnd ='1'; f_bG =1; }               //default '1'
                            break;
  b_WhiskerReact: case 'W': if (f_recording) { RecordCcmnd(); }    //so 'W' is recorded
                            GetCcmnd();                            //get parameter           
                            if (Ccmnd=='0') { f_bW =0; } 
                            else 
                            { Ccmnd ='1'; f_bW =1; }               //default '1'
                            break;
  b_SonarReact:   case 'U': if (f_recording) { RecordCcmnd(); }    //so 'U' is recorded
                            GetCcmnd();                            //get parameter           
                            if (Ccmnd=='0') { f_bU =0; } 
                            else 
                            { Ccmnd ='1'; f_bU =1; }               //default '1'
                            break;
  b_AutoSense:    case '$': if (f_recording) { RecordCcmnd(); }    //so '$' is recorded
                            GetCcmnd();                            //get parameter           
                            if (Ccmnd=='1') { f_bAutoSense =1; }
                            else 
                            { Ccmnd ='0'; f_bAutoSense =0; }       //default '0'   
                            break;
  b_verbose:      case 'v': if (f_recording) { RecordCcmnd(); }    //so 'v' is recorded          
                            GetCcmnd();                            //get parameter           
                            if (Ccmnd=='1') { f_bv =1; }
                            else 
                            { Ccmnd ='0'; f_bv =0; }               //default '0'   
                            break;
  b_ServoAdjust:  case '@': if( f_BeQuiet==1 ) { break; }     // don't adjust if being quiet
                            AdjustServos();  break;           
  b_default:      default:  badkey();  if(!f_BeQuiet){Serial.println("bad key");}
                            Ccmnd ='?';     //so if recording then '?' is recorded
                            break;          //don't do anything
                }     // end switch
                break;                               
//------------end case 'b'

SYSTEM:  
    case 'r':   f_recordC =false;
                redLEDon();
                if (!f_recording) 
                  { getAct();  if(Ccmnd=='.') { break; }  
                  setEEPROMindexes(Act);  f_recording =true; 
                  }
                break;
    case 'c':   f_recordC =false;  if (!f_recording) { copyAct(); }  break;	
    case 'l':   f_recordC =false;  if (!f_recording) { loadAct(); }  break;	
    case '?':   f_recordC =false;
                if (!f_recording) 
                  { GetCcmnd();                           //get parameter
                    if( f_BeQuiet==1 ) { break; }         // don't print if being quiet
                    switch (Ccmnd)
                    { case 'h':  printHelp();  break;     // print instructions
                      case '1':  // fall through
                      case '2':  // fall through
                      case '3':  // fall through
                      case '4':  // fall through
                      case '5':  // fall through
                      case '6':  // fall through
                      case '7':  // fall through
                      case '8':  Act =Ccmnd;  printAct();  break;
                      case 'p':  PrintAllActs();  break;
                      case '?':  PrintAllActs();  break;
                      case 'd':  DumpActs();  break;
                      case 'b':  printBehaviour();  break;
                      case 'm':  printMind();  break;      // - all Acts, Behaviour
                      case '!':  PrintReActs(); break;
                      case 'B':  PrintBehaviours(); break;
                      case '$':  readUSdistance(); printIRsensors();  break;
                      default:   badkey();  if(!f_BeQuiet){Serial.println("?");}  break;
                    }    // end switch               
                  }    // end if 
                break;
    case '#':   f_recordC =false;
                f_MsgForMe =1;              // so next GetCcmnd() works
                f_hashcmnd =true;           // don't get a Play command here
                GetCcmnd();                 // get parameter
                f_hashcmnd =false;
                switch (Ccmnd)
                { case '0': f_MsgForMe =1;  f_BeQuiet =1;  hashgroup =rgroup;  break;  //everybody listen
                  case '#': f_MsgForMe =1;  f_BeQuiet =0;                  //everybody stop
                            f_hashwalking =f_walking;
                            disable_walking();  Ccmnd ='.';  hashhashAt =EEPROMindex;
                            f_hashhashSTOP =true;  // so '.' gets processed
                            f_hashrunDefault =f_runDefault;  f_runDefault =false;
                            f_hashbAutoSense =f_bAutoSense;  f_bAutoSense =false;
                            break;
                  case '<': f_doDefault =true;  f_hashhashSTOP =false;  
                            f_runDefault =f_hashrunDefault;
                            f_bAutoSense =f_hashbAutoSense;
                            enable_walking();
                            break;
                  case '>': f_runDefault =f_hashrunDefault;
                            f_bAutoSense =f_hashbAutoSense;
                            if (hashhashAt!=-1)                //something to continue
                              { f_play =true;  EEPROMindex =hashhashAt; }
                            f_hashhashSTOP =false;    
                            enable_walking();
                            break;
                  case '?': delay((rgroup -'0')*50+(rname-'A')*10);
                            Serial.print(" "); Serial.print(rgroup); Serial.print(rname); 
                            break;
				          case '.':	f_MsgForMe =0;  f_BeQuiet =1;  hashgroup =' ';       //end #session
				                    break;
                  default:  if (Ccmnd>='1' && Ccmnd<='9' ) {
                              hashgroup =Ccmnd; 
                              if (hashgroup==rgroup) 
                                { f_MsgForMe =1;  f_BeQuiet =1;  }    //my group
                               else
                                { f_MsgForMe =0;  f_BeQuiet =1;  }    //not my group
                              break;
                              }
                            if (Ccmnd==rname && hashgroup==rgroup)
                              { f_MsgForMe =1;  f_BeQuiet =0;                 //me
                                if(hashgroup==0) { f_BeQuiet =1; }  break; }  //me and others
                             else
                              { f_MsgForMe =0;  f_BeQuiet =1;  break; }       //not me 
                            break;
                }
                break;      
SYSTEM_InitMind:
    case '*': if (f_recording) { break; }        //no action if recording
              GetCcmnd(); if (Ccmnd=='i') { GetCcmnd(); if (Ccmnd=='M') { initialiseEEPROM(); } }
              break;
    // end SYSTEM  
COMMAND_Default:                
    default:    f_recordC =false;    //don't record bad key		
                badkey();  if(!f_BeQuiet){ Serial.println(" bad key"); }  break;
  }    // end switch COMMAND
	
Record:	
  if (f_recording  && f_recordC) { RecordCcmnd(); } 
  
}
//== END {LOOP} =============================================

//== USER ROUTINES ==========================================
void Experiment()       //until Quit
{ while(true)
  { 
    //put your own routine here e.g.
    rollspeed =4;
    pacespeed =4;
    forward();       // Walk one step forwards
    forward();       // Walk one step forwards
    forward();       // Walk one step forwards
    whee(); 
    turnRight();
    turnRight();
    turnRight();    
    whee(); 
    rollspeed =2;
    pacespeed =2;
    turnLeft();
    turnLeft();
    turnLeft();
    whee(); 
    rollspeed =1;
    pacespeed =1;
    backward();      // Walk one step backwards
    backward();      // Walk one step backwards
    backward();      // Walk one step backwards
    whee();    
    whee();    
    whee();    
    if(Quit()) { break; } 
  }	// end while
}
//----------------------------
void Xplore()         //Wander() until Quit
{ while(true)
  { Wander();
    if(Quit()) { break; } 
  }	// end while
}
//----------------------------
void beAlive()         // the loop() from B_PM22_BeAlive
{ while(true)
  { switch (random(9))
    { case 0: WanderSteps(random(10)); break;    //uses all sensors
      case 1: rock(random(3)); break;
      case 2: shuffle(random(3)); break;
      case 3: whoa(); break;
      case 4: whee(); break;
      case 5: rollspeed =random(1,4); break;
      case 6: pacespeed =random(1,6); break;
      case 7: RightStep(random(6)); break;
      case 8: LeftStep(random(6)); break;
    } // end switch
    if( Quit() ) { break; }
  }	// end while
}

//== FUNCTIONS ==============================================

void GetCcmnd()	       //reads Serial or playing Act until Ccmnd != 0, prints Ccmnd
{ Ccmnd =0;
  while(Ccmnd==0)
  { if (f_recording)  
    { //while(!Serial.available()) {}       //wait for Serial
      if (IRserial.available())
      { Ccmnd =IRserial.read();  bip(); }          //IR commands are global, override everthing if not recording 
      else 
      if (Serial.available()) 
      { Ccmnd =Serial.read();  bip(); }
    } 
    else         //if (!f_recording)
    { if (IRserial.available())
      { Ccmnd =IRserial.read();  bip(); }          //IR commands are global, override everything if not recording 
      else 
      { if (Serial.available()) 
        { Ccmnd =Serial.read();                    //over-ride playing if Serial, null ignored
          if (!f_MsgForMe && Ccmnd!='#') 
          { Ccmnd =0; }     //discard messages for others
          else 
         { bip(); }
        }          
        else // are we playong or reacting
        { if (!f_hashcmnd)         //don't react or get a Play character if in the middle of a hash command   
          { if (f_play || f_reacting)
            { Ccmnd =PlayAct(); }            //get Act or React Ccmnd 
          }  
        }  //end if (!f_hashcmnd)
      } //end of if (Serial.available())
    }  //end if (!f_recording)                                               
  }  //end while Ccmnd==0
  printCcmnd();
}   
//----------------------------
void printCcmnd()
{ if (!f_BeQuiet && !f_skipmind){ Serial.print(Ccmnd); }
} 
//----------------------------
void RecordCcmnd()	
{ EEPROM.write(EEPROMindex,Ccmnd);
  EEPROMindex++;
  if (EEPROMindex>=ActEnd) { f_recording =false;  EEPROM.write(ActEnd,'.');  badkey(); }     //terminated string 
}  
//----------------------------
void makeCcmndDigit()      //ensure 0 - 9
{ if (Ccmnd<'0' || Ccmnd>'9') { Ccmnd ='0'; } //default
}
//----------------------------  
void waitCcmndSecs()   //read delay time from Act and wait
{ if (Ccmnd<'1' || Ccmnd>'9') { Ccmnd ='1'; } //default
  byte cmnd =Ccmnd -'0';                          // ASCII to numeric
  delay(cmnd *1000);
}  
//----------------------------  
void printMind()
{ PrintAllActs();
  printBehaviour();
}
//----------------------------
void PrintReActs()
{ Serial.println();  Serial.print("W & G, 1 indicates object or hole sensed LR");
  for(Act ='1'; Act<='8'; Act++) { epage =3; printReAct(); }  Serial.println(); 
}
//----------------------------
void printReAct()
{ Serial.println();
  Serial.print("ReAct");  Serial.print(Act);  Serial.print(" ");
  switch (Act)
  { case '1':  Serial.print("U<=");  Serial.print(SonarAwareAt);  Serial.print(" ");  break;
    case '2':  Serial.print("U<=");  Serial.print(SonarTooNear);  Serial.print(" ");  break;
    case '3':  Serial.print("W11");  Serial.print("  ");  break;
    case '4':  Serial.print("W10");  Serial.print("  ");  break;
    case '5':  Serial.print("W01");  Serial.print("  ");  break;
    case '6':  Serial.print("G11");  Serial.print("  ");  break;
    case '7':  Serial.print("G10");  Serial.print("  ");  break;
    case '8':  Serial.print("G01");  Serial.print("  ");  break;
  } // end switch
  setEEPROMindexes(Act);
  Ccmnd =EEPROM.read(EEPROMindex++);
  // -1 (=255) is uninitialised EEPROM value
  while(Ccmnd !='.' && Ccmnd !=-1) { Serial.print(Ccmnd); moves++; Ccmnd =EEPROM.read(EEPROMindex++); }
  EEPROMindex -=1;        //step back to end of record just in case playing
  Serial.print(". - ");
  Serial.print(moves);    //there are moves which take more than 1 character  
  Serial.print(" bytes"); //so the space taken in EEPROM is really bytes
  moves =0;	
}
//---------------------------- 
void loadBehaviour()
{ getBehaviour();  // 1 - 9   '.' quits
  Serial.print("Behaviour="); if(Behaviour=='9') { Serial.print("all"); } else { Serial.print(Behaviour); } 
  Serial.println("  , ; [Enter]   to enter numbers, [Esc] or '.' to quit");
  while( Ccmnd!='.' )    
  { Serial.println("enter - behaviour_number, behaviour_value,");
    do                                  //get BeHave address in the slot, ended with ','
    { Ccmnd =' ';
      if(Serial.available()) { Ccmnd =Serial.read(); Serial.print(Ccmnd);  bip(); }
      if (isDigit(Ccmnd)) { numString += Ccmnd; Ccmnd =' '; }
       //[Return] is \r \n so let isDigit filter \r      
      }while( Ccmnd!='\n' && Ccmnd!=',' && Ccmnd!='.' && Ccmnd!=Esc ); 
      
    if(Ccmnd=='.' || Ccmnd==Esc) break;             //exit while
    behaveNum =numString.toInt(); numString ="";
Serial.print("behaveNum=");Serial.println(behaveNum);

    switch (behaveNum)
    { case 31:  //Ef_walking    - true, false
                get1char(); behaveValue =constrain(Ccmnd -'0',0,1);
                break;
      case 30:  //ErollC        - ~90
                getnum();   behaveValue =constrain(behaveValue,80,100);  
                break;
      case 29:  //EpaceC        - ~90
                getnum();   behaveValue =constrain(behaveValue,80,100);  
                break;
      case 28:  //Erollby       - 1-15
                getnum();   behaveValue =constrain(behaveValue,1,maxroll);  
                break;
      case 27:  //Epaceby       - 1-15 
                getnum();   behaveValue =constrain(behaveValue,1,maxpace);  
                break;
      case 22:  //Ergroup         - '1-9'
                get1char();  behaveValue =constrain(Ccmnd,'1','9');
                break;
      case 21:  //Ername        - 'A-z'
                get1char();  behaveValue =constrain(Ccmnd,'A','z');
                break;
      case 20:  //EautoAct      - '0-8'
                get1char();  behaveValue =constrain(Ccmnd,'0','8');
                break;
      case 19:  //ESonarAwareAt - 1-9
                get1char();  behaveValue =constrain(Ccmnd -'0',1,9);
                break;
      case 18:  //ESonarTooNear - 1-9
                get1char(); behaveValue =constrain(Ccmnd -'0',1,9);
                break;
      case 17:  //Ef_cm         - true, false
                get1char();  behaveValue =constrain(Ccmnd -'0',0,1);
                break;
      case 16:  //Erollspeedmax - 1-9
                get1char();  behaveValue =constrain(Ccmnd -'0',1,9);
                break;
      case 15:  //Epacespeedmax - 1-9
                getchar();  behaveValue =constrain(Ccmnd -'0',1,9);
                break;
      case 14:  //Erollspeed    - 1-9
                get1char();  behaveValue =constrain(Ccmnd -'0',1,9);
                break;
      case 13:  //Epacespeed    - 1-9
                get1char();  behaveValue =constrain(Ccmnd -'0',1,9);
                break;
      case 12:  //f_bG    - true, false
                get1char();  behaveValue =constrain(Ccmnd -'0',0,1);
                break;
      case 11:  //f_bW    - true, false
                get1char();  behaveValue =constrain(Ccmnd -'0',0,1);
                break;
      case 10:  //f_bU    - true, false
                get1char();  behaveValue =constrain(Ccmnd -'0',0,1);
                break;
      case 9:   //f_AutoSense    - true, false
                get1char();  behaveValue =constrain(Ccmnd -'0',0,1);
                break;
      default:  behaveNum =0;
                Serial.print("bad BeHave number");  Serial.println();  
                break;      
      }   //end switch

Serial.print("behaveValue=");Serial.println(behaveValue);
      
    if(behaveNum!=0) { 
      if(Behaviour!='9') { 
        BehaviourEbase =(Behaviour -'1') *32; 
        EEPROM.write(BehaviourEbase +behaveNum, behaveValue);
       }   
       else {
         for ( Behaviour =0; Behaviour <8; Behaviour ++ )  {
           BehaviourEbase =Behaviour *32; 
           EEPROM.write(BehaviourEbase +behaveNum, behaveValue);
          }
         Behaviour ='9'; 
        }
      }  
    } //end while
}
//---------------------------- 
void get1char()
{ Ccmnd =0;
  do{ if(Serial.available())  Ccmnd =Serial.read(); }while(Ccmnd==0);
  Serial.println(Ccmnd);  bip();
}
//----------------------------  
void getnum()
{ do
  { Ccmnd =' ';
    if(Serial.available()) { Ccmnd =Serial.read();  Serial.print(Ccmnd);  bip();} 
    if (isDigit(Ccmnd)) { numString +=Ccmnd;   Ccmnd =' '; }
       //[Return] is \r \n so let isDigit discard \r      
    }while(Ccmnd!='\n' && Ccmnd!='.' && Ccmnd!=',' && Ccmnd!=';' && Ccmnd!=Esc);
  behaveValue =numString.toInt();  numString = "";
  if(Ccmnd=='.' || Ccmnd==Esc) { Serial.println();  behaveNum =0; }
}
//----------------------------  
char getBehaviour()      //called when loading a behaviour, Ccmnd ->'.', '1' - '8'
{ Ccmnd =0;
  do
  { if(Serial.available()) 
    { Ccmnd =Serial.read(); bip();
      switch (Ccmnd) 
      { case Esc:  Ccmnd ='.';     //[Esc] key, fall through to '.'
        case '.':  break;           //quit
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':  Behaviour =Ccmnd;  break;
        default:   badkey(); Serial.print(" enter Behaviour or '.'"); Ccmnd =0;
      }  //end case
    } //end if Serial
  } while(Ccmnd==0);
  Serial.println(Ccmnd);
}    
//----------------------------  
void printBehaviour()
{ Serial.println();                
  Serial.println(thisprog); // so we know which program is running
  Serial.print("robot ");      Serial.print(rgroup);  Serial.print(rname);
                               Serial.print(" MsgForMe=");  Serial.print(f_MsgForMe); 
                               Serial.print(" BeQuiet=");   Serial.println(f_BeQuiet); 
  Serial.print("b  "); Serial.print(Behaviour);     Serial.println(" Behaviour"); 
  Serial.print("bw "); Serial.print(f_walking);     Serial.println(" walking"); 
  Serial.print("ba "); Serial.print(autoAct);       Serial.print(" autoAct,  "); Serial.print("rerun "); 
      if (f_runDefault==1) { Serial.print("On"); } else { Serial.print("Off"); } Serial.println("  (default Off)"); 
  Serial.print("sr "); Serial.print(rollspeed);  Serial.print(',');  Serial.print(rollspeedmax);  Serial.println(" rollspeed,rollspeedmax");
  Serial.print("sp "); Serial.print(pacespeed);  Serial.print(',');  Serial.print(pacespeedmax);  Serial.println(" pacespeed,pacespeedmax");
  Serial.print("br "); Serial.print(rollby/rollbyScale);  Serial.print('(');  Serial.print(rollby);  Serial.print(')');  Serial.println(" rollby");   
  Serial.print("bp "); Serial.print(paceby/pacebyScale);   Serial.print('(');  Serial.print(paceby);  Serial.print(')');  Serial.println(" paceby");
  Serial.print("bt "); Serial.print(turnby/turnbyScale);  Serial.print('(');  Serial.print(turnby);  Serial.print(')');  Serial.println(" turnby");
  Serial.print("br+- "); Serial.print(rollC);    Serial.println(" rollC");    
  Serial.print("bp+- "); Serial.print(paceC);    Serial.println(" paceC");
  Serial.print("bu+ "); Serial.print(SonarAwareAt);  Serial.println(" SonarAwareAt");   
  Serial.print("bu- "); Serial.print(SonarTooNear);  Serial.println(" SonarTooNear");   
  Serial.print("bG "); Serial.print(f_bG);          Serial.println(" Greact");   
  Serial.print("bW "); Serial.print(f_bW);          Serial.println(" Wreact");   
  Serial.print("bU "); Serial.print(f_bU);          Serial.println(" Ureact");   
  Serial.print("b$ "); Serial.print(f_bAutoSense);  Serial.println(" auto$ense");
  Serial.print("bv "); Serial.print(f_bv);          Serial.println(" verbose");
//disp_freeram();  
}
//----------------------------  
void PrintBehaviours()
//This uses Serial.print(F(" text ")); which saves RAM by leaving the text in Program memory 
//and is a lot easier to use than the PROGMEM construct previously used for the help text - we can learn something every day!
{ Serial.println();
  Serial.println(F("BEHAVIOUR       b1    b2    b3    b4    b5    b6    b7    b8"));
  Serial.println(F(" # BeHave       --    --    --    --    --    --    --    --"));
  Serial.print(F("31 f_walking    "));
    for(int i =31; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("     ")); } Serial.println("0=false, 1=true");
  Serial.print(F("30 rollC        "));
    for(int i =30; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("    ")); } Serial.println("servo degrees");
  Serial.print(F("29 paceC        "));
    for(int i =29; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("    ")); } Serial.println("servo degrees");
  Serial.print(F("28 rollby       "));
    for(int i =28; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("    ")); } Serial.print("degrees =cmnd *"); Serial.println(rollbyScale);
  Serial.print(F("27 paceby       "));
    for(int i =27; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("    ")); } Serial.print("degrees =cmnd *"); Serial.println(pacebyScale);
  //23-26 never altered
  Serial.print(F("22 rgroup       "));
    for(int i =22; i<256; i +=32) { Serial.print(char(EEPROM.read(i))); Serial.print(F("     ")); } Serial.println("1-9");
  Serial.print(F("21 rname        "));
    for(int i =21; i<256; i +=32) { Serial.print(char(EEPROM.read(i))); Serial.print(F("     ")); } Serial.println("A-z");
  Serial.print(F("20 autoAct      "));
    for(int i =20; i<256; i +=32) { Serial.print(char(EEPROM.read(i))); Serial.print(F("     ")); } Serial.println("0-8");
  Serial.print(F("19 SonarAwareAt "));
    for(int i =19; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("     ")); } Serial.println("0-9, units 5cm or 2inch");
  Serial.print(F("18 SonarTooNear "));
    for(int i =18; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("     ")); } Serial.println("0-9, units 5cm or 2inch");
  Serial.print(F("17 Sonar use cm "));
    for(int i =17; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("     ")); } Serial.println("0=false, 1=true");
  Serial.print(F("16 rollspeedmax "));
    for(int i =16; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("     ")); } Serial.println("1-9");
  Serial.print(F("15 pacespeedmax "));
    for(int i =15; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("     ")); } Serial.println("1-9");
  Serial.print(F("14 rollspeed    "));
    for(int i =14; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("     ")); } Serial.println("1-9");
  Serial.print(F("13 pacespeed    "));
    for(int i =13; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("     ")); } Serial.println("1-9");
  Serial.print(F("12 Greact       "));
    for(int i =12; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("     ")); } Serial.println("0,1");
  Serial.print(F("11 Wreact       ")); 
    for(int i =11; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("     ")); } Serial.println("0,1");
  Serial.print(F("10 Ureact       "));   
    for(int i =10; i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("     ")); } Serial.println("0,1");
  Serial.print(F(" 9 auto$ense    "));   
    for(int i =9;  i<256; i +=32) { Serial.print(EEPROM.read(i));       Serial.print(F("     ")); } Serial.println("0,1");
  Serial.print(F("-- @Behaviour   "));
    for(int i =0; i<256; i +=32) { Serial.print(char(EEPROM.read(i)));  Serial.print(F("     ")); } Serial.println();
} 
//----------------------------  
void DumpActs()        //Dump Acts, Acts0, ReActs. Print without names and length
{ for(Act ='1'; Act<='8'; Act++)
  { setEEPROMindexes(Act);
    Serial.println();
    Ccmnd =EEPROM.read(EEPROMindex++);
    // -1 (=255) is uninitialised EEPROM value
    while(Ccmnd !='.' && Ccmnd !=-1) { Serial.print(Ccmnd); Ccmnd =EEPROM.read(EEPROMindex++); }
  }
  for(Act ='1'; Act<='8'; Act++) 
  { epage =2;  setEEPROMindexes(Act);
    Serial.println();
    Ccmnd =EEPROM.read(EEPROMindex++);
    // -1 (=255) is uninitialised EEPROM value
    while(Ccmnd !='.' && Ccmnd !=-1) { Serial.print(Ccmnd); Ccmnd =EEPROM.read(EEPROMindex++); }
  }
  for(Act ='1'; Act<='8'; Act++) 
  { epage =3;  setEEPROMindexes(Act);
    Serial.println();
    Ccmnd =EEPROM.read(EEPROMindex++);
    // -1 (=255) is uninitialised EEPROM value
    while(Ccmnd !='.' && Ccmnd !=-1) { Serial.print(Ccmnd); Ccmnd =EEPROM.read(EEPROMindex++); }
  }
  Serial.println();
}		
//----------------------------  
void loadAct()
{ redLEDon();  
  getAct();                         // Ccmnd ->'.' - Quit,
                                    // '0', '!'    - sets Acts0 or ReActs page then gets Act
                                    // '1' - '8'   - sets Act '1' - '8',  - 
  if (Ccmnd!='.')
  { setEEPROMindexes(Act);
    while(true) 
    { while(!Serial.available()) {}  //wait for Serial
      Ccmnd =Serial.read();  Serial.print(Ccmnd);
      if(Ccmnd=='.' || Ccmnd==Esc) { break; }
      EEPROM.write(EEPROMindex,Ccmnd);  EEPROMindex++;
      if (EEPROMindex>=ActEnd) { badkey();  break; }     //ActEnd =EEPROMindex +31;   
    }
    EEPROM.write(EEPROMindex,'.');                        //terminated string 
  }
}
//----------------------------  
void copyAct()
{ getAct();
  if (Ccmnd!='.')
  { setEEPROMindexes(Act);
    int fromEEPROMindex =EEPROMindex;  
    int fromActEnd =ActEnd;
    getAct();
    if (Ccmnd!='.')
    { setEEPROMindexes(Act);
      for(byte i =0; i<=fromActEnd -fromEEPROMindex; i++)
      { EEPROM.write(EEPROMindex +i,EEPROM.read(fromEEPROMindex +i)); }
    } 
  }	
}
//----------------------------
char PlayAct() 
{ char _Ccmnd =EEPROM.read(EEPROMindex++);
  if(_Ccmnd==255 || _Ccmnd==-1) _Ccmnd ='.';     //uninitialised EEPROM value, don't play it, change to '.'
  return _Ccmnd;
}  
//----------------------------
void PrintAllActs()
{ for(Act ='1'; Act<='8'; Act++) { epage =1; printAct(); }  Serial.println(); 
  for(Act ='1'; Act<='8'; Act++) { epage =2; printAct(); }  Serial.println(); 
}
//----------------------------
void printAct()
{ Serial.println();
  Serial.print("Act");
  if(epage==2) { Serial.print('0'); }    //for logical page0 Acts
  Serial.print(Act);
  Serial.print(" ");
  setEEPROMindexes(Act);
  Ccmnd =EEPROM.read(EEPROMindex++);
  // -1 (=255) is uninitialised EEPROM value
  while(Ccmnd !='.' && Ccmnd !=-1) { Serial.print(Ccmnd); moves++; Ccmnd =EEPROM.read(EEPROMindex++); }
  EEPROMindex -=1;        //step back to end of record just in case playing
  Serial.print(". - ");
  Serial.print(moves);
  Serial.print(" bytes");
  moves =0;	
}
//----------------------------  
void getAct()      //called when recording, loading, printing and copying Acts, Ccmnd ->'.', '1' - '8'
{ Ccmnd =0;
  do
  { if(Serial.available())    { Ccmnd =Serial.read(); Serial.print(Ccmnd); }
    if(IRserial.available())  { Ccmnd =IRserial.read(); bip(); }
    if (Ccmnd!=0)
      {
      switch (Ccmnd) 
      { case Esc:   Ccmnd ='.';     //[Esc] key, fall through to '.'
        case '.':  Serial.print(">");  break;     //quit, print ready prompt
        case '!':  epage =3;  Ccmnd =0;  break;  //do it again to read Act
        case '0':  epage =2;  Ccmnd =0;  break;  //do it again to read Act
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':  Act =Ccmnd;  break;
        default: badkey(); Serial.print(" enter Act or '.'"); Ccmnd =0;	
      }  //end case
    } //end if Serial
  } while(Ccmnd==0);
}    
//----------------------------
void setEEPROMindexes(char Act)     // set EEPROM pointers
{ actIndex =Act -'1';               // ASCII to numeric,  Act 1 == actIndex 0
  EEPROMindex =(256 *epage) +(actIndex*32);  
  ActEnd =EEPROMindex +31;
  epage =1;            //set to default page, Acts 1 - 8 
}
//----------------------------
void badkey()     // indicate invalid key
{ digitalWrite(RwhiskerLEDpin_, LOW);// whisker LED ON
  delay(100);
  digitalWrite(RwhiskerLEDpin_, HIGH);// whisker LED OFF    
  digitalWrite(LwhiskerLEDpin_, LOW);// whisker LED ON
  delay(100);
  digitalWrite(LwhiskerLEDpin_, HIGH);// whisker LED OFF  
  whoa();  
}  
//----------------------------
void USreact()     
// "U" read Ultrasonic sensor and react to objectRange
{ readUSdistance();
  if (f_bU) { reactUObjects(); }
}
//----------------------------
void WhiskerReact()     
// "W" read Whisker sensors and avoid obstacles (if any) using avoidObstacles() behaviour rules
{ 
#ifdef BambinoMM
  return;
#endif  
  readWhiskers();
  queryObstaclesLR();
  if (f_bW) { reactObstaclesLR(); }
}
//----------------------------
void BeardReact()     
// "G" Ground - read Beard sensors and avoid holes (if any) using avoidHoles() behaviour rules
{ readBeard();
  queryEdgeLR();
  if (f_bG) { //reactHolesLR(); }
    delay(20);
    readBeard();
    queryEdgeLR();
    if (f_bG) { reactHolesLR(); }
  }
}
//---------------------------
boolean Quit()          //check for Quit
{ if( Serial.available() ) 
  { Ccmnd =Serial.peek(); 
    if( Ccmnd=='.' || Ccmnd=='Q' || Ccmnd==Esc || Ccmnd=='#')
      { return true; }
     else
      { return false; }
  }  
}
//----------------------------
void bip()            // make a bip sound
{ beep(voicePin,1500,50); }
//----------------------------
#ifdef BambinoMM
void redLEDdim()
{ analogWrite(redLEDpin_,254); }  //redLED, LOW=on so 255 is off 
//----------------------------
void redLEDon()
{ digitalWrite(redLEDpin_,LOW); }  
//----------------------------
void redLEDoff()
{ digitalWrite(redLEDpin_,HIGH); }  
//---------------------------- 
//BambinoMM does not have Whisker sensors
//---------------------------- 
#else
void redLEDon()
{ digitalWrite(redLEDpin,HIGH); }  
//----------------------------
void redLEDoff()
{ digitalWrite(redLEDpin,LOW); }  
//----------------------------
void whiskerLEDsOff()
{ digitalWrite(RwhiskerLEDpin_, HIGH);     // whisker LED OFF
  digitalWrite(LwhiskerLEDpin_, HIGH);     // whisker LED OFF
  }  
//----------------------------  
void whiskerLEDsDim()
{ analogWrite(RwhiskerLEDpin_,254);   //Right whiskerLED, LOW=on so 255 is off
  analogWrite(LwhiskerLEDpin_,254);   //Left whiskerLED, LOW=on so 255 is off
  }
#endif
//----------------------------  
void beardLEDsOff()
{ digitalWrite(RbeardLEDpin_, HIGH);       // beard LED Off
	digitalWrite(LbeardLEDpin_, HIGH);       // beard LED Off
  }
//===========================================================
/*****************Dan*************************/
// print integer value. 
/*void i_printf( int val ) {  Serial.print(val);  }

// print string s1. 
void s_printf( char *s1 ) {  Serial.print(s1);  }
  
/*********************************************/
/*void disp_freeram()
{
  s_printf("-SRAM left: ");
  i_printf( freeRam() );
}
/*********************************************/
/*int freeRam() 
{
  extern int __heap_start,*__brkval;
  int v;
  return (int)&v - (__brkval == 0  
    ? (int)&__heap_start : (int) __brkval);  
}  
/*********************************************/
